"""The actual command classes"""

import re
from typing import Union, List, Callable, Any, Tuple, Dict, Optional
from argparse import ArgumentParser

def _def_callback(cmd: Command, args: List[str]) -> None:
    """Default callback for any command (if no other callback is provided)."""
    pass

class Command(object):
    """
    Represents an actual command, something the user can enter into the shell.

    ...

    Attributes
    ----------
    aliases : List[str]
        A list of all the keywords (aliases) that the command has; that will trigger it.
    case_sensitive : bool
        Whether or not the command should respond to both the upper- and lowercase 
        versions of its aliases.
    callback : Callable[..., None]
        The function that will be called, when the command is executed.
    help : str
        A short summary of what the command does.
    cbargs : Union[Tuple[()], Tuple[Any]]
        Positional arguments that will be passed to the callback function.
    cbkwargs : Dict[str, Any]
        Keyword arguments that will be passed to the callback function.
    parent : Command
        The command's supercommand (needed for tracing).

    Methods
    -------
    matches(args)
        Checks whether or not the command should be triggered by the given cl arguments.
    trace(cu=None)
        Computes the command trace. The chain of supercommands that is needed to trigger this command.
    usage()
        Returns a short summary of this commands arguments and trace.
    __call__()
        Will call the specified callback function.
    __str__()
        Returns a nicely formatted, short help string for the command.
    """
    def __init__(self, cmd: str, 
                       *aliases: str, 
                       case_sensitive: bool = True, 
                       callback: Callable[..., None] = _def_callback, 
                       hint: str = '', 
                       cbargs: Union[Tuple[()], Tuple[Any]] = (), 
                       cbkwargs: Optional[Dict[str, Any]] = None, 
                       parent: Optional["Command"] = None
                ):
        """
        Parameters
        ----------
        cmd : str
            The main keyword that will trigger this commands.
        *aliases : str
            Aliases for the main keyword.
        case_sensitive : bool
            Whether or not the lower-, upper- and mixedcase versions of the aliases should be acceptable.
            Default is True.
        callback : Callable[..., None]
            The function that will be called when this command is triggerd.
            Default is _def_callback.
        hint : str
            The short description (will be help)
            Default is ''.
        cbargs : Union[Tuple[()], Tuple[Any]]
            A tuple of positional arguments that will be passed to the callback.
            Default is ().
        cbkwargs : Optional[Dict[str, Any]]
            A dictionary of keyword arguments that will be passed to the callback.
            Default is None.
        parent : Optional[Command]
            The command's supercommand.
            Default is None.
        """
        self.aliases: List[str] = [cmd, *aliases]
        self.case_sensitive: bool = case_sensitive
        self.callback: Callable[..., None] = callback
        self.help: str = hint
        self.cbargs: Union[Tuple[()], Tuple[Any]] = cbargs
        self.cbkwargs: Dict[str, Any] = cbkwargs or dict()
        self.parent: "Command" = parent
        self.parser = ArgumentParser()

    def add_arg(self, *args, **kwargs) -> None:
        self.parser.add_argument(*args, **kwargs)

    def matches(self, args: Union[str, List[str]]) -> bool:
        """
        Checks whether or not the command should be triggered by the given cl arguments.
        
        ...

        Parameters
        ----------
        args : Union[str, List[str]]
            Either the commandline or the already split arguments.

        Returns
        -------
        bool
            Whether or not the command should be executed.
        """
        cmd: str = (args[0] if isinstance(args, list) else args.split()[0]).strip()
        return cmd in self.aliases or (cmd.lower() in [a.lower() for a in self.aliases] and not self.case_sensitive)

    def trace(self, cu: Optional["Command"] = None) -> str:
        """
        Will compute and return the command's trace.

        Recursively goes through the supercommands until it reaches the shell;
        returns the resulting trace.

        Parameters
        ----------
        cu : Optional[Command]
            Since it's recursive - the current command.
            Default is None

        Returns
        -------
        str
            The trace (chain of keywords) needed to trigger the command.
        """
        c = cu
        if not c:
            c = self
        return (self.trace(c.parent) if c.parent else '') + c.aliases[0] + (' ' if cu else '')

    def usage(self) -> str:
        """Will return how to use the command."""
        return self.trace() + ' ' + self.parser.format_help()[self.parser.format_help().index('['):]

    def __call__(self, args: Union[str, List[str]]) -> None:
        """Will call the callback function."""
        try:
            ar = self.parser.parse_args(args if isinstance(args, list) else args.split())
        except SystemExit:
            print(self.usage())
            return
        self.callback(self, args if isinstance(args, list) else args.split(), *self.cbargs, **self.cbkwargs, **vars(ar))

    def __str__(self) -> str:
        """Returns a nicely formatted help-string."""
        return self.aliases[0] + '\t' + self.help

def _def_unkown_key(cc: CascCommand, cmdline: str) -> None:
    """The default callback for an unknown key."""
    print('Usage: %s' % cc.usage())

class CascCommand(Command):
    """
    Represents a cascading-command; a command with sub-commands.

    ...

    Attributes
    ----------
    cmds : List[Command]
        A list of all availabe sub-commands.
    sep : str
        The argument separator.
    unknown_key : Callable[[CascCommand, str], None]
        The function that will be called, if an unknown sub-command is encountered.

    Methods
    -------
    parse(cmdline)
        Parses the cmdline; checks which sub-command should be called. 
    add_cmd(cmd)
        Adds a command to its list of sub-commands.
    usage()
        Will return a trace to the command + all sub-commands summarized.
    """
    def __init__(self, cmd: str,
                       *aliases: str,
                       cmds: Optional[List[Command]] = None, 
                       case_sensitive: bool = True,
                       unknown_key: Callable[["CascCommand", str], None] = _def_unkown_key, 
                       callback: Callable[..., None] = _def_unkown_key, 
                       hint: str = '',
                       cbargs: Union[Tuple[()], Tuple[Any]] = (), 
                       cbkwargs: Optional[Dict[str, Any]] = None, 
                       parent: Optional[Command] = None,
                       sep: str = r'\s(?:(?=(?:[^"]*"[^"]*")+[^"]*$)|(?=[^"]*$))',
                ):
        """
        Parameters
        ----------
        cmd : str
            The main keyword to trigger this command.
        *aliases : str
            The main keyword's aliases; will also trigger the command.
        cmds : Optional[List[Command]]
            A list of all available sub-commands.
            Default is None.
        case_sensitive : bool
            Whether or not the keywords should be case-sensitive.
            Default is True.
        unknown_key : Callable[[CascCommand, str], None]
            The function to-be-called if an unknown sub-command is encountered.
            Default is _def_unknown_key.
        callback : Callable[..., None]
            The function that should be called upon the command's execution.
            Default is _def_unknown_key.
        hint : str
            A short summary of the command's functionality & purpose.
            Default is ''.
        cbargs : Union[Tuple[()], Tuple[Any]]
            Positional arguments that will be passed to the callback.
            Default is ().
        cbkwargs : Optional[Dict[str, Any]]
            Keyword arguments that will be passed to the callback.
            Default is None.
        parent : Optional[Command]
            The parent of the casc-command; its supercommand.
            Default is None.
        sep : str
            The argument separator.
            Default is r'\s(?:(?=(?:[^"]*"[^"]*")+[^"]*$)|(?=[^"]*$))'.
        """
        super().__init__(cmd, *aliases, callback=callback, case_sensitive=case_sensitive, hint=hint, cbargs=cbargs, cbkwargs=cbkwargs, parent=parent)
        self.cmds: List[Command] = cmds or []
        self.unknown_key: Callable[["CascCommand", str], None] = unknown_key
        self.sep: str = sep
        for c in self.cmds:
            c.parent = self

    def parse(self, cmdline: str) -> None:
        """
        Parses the cmdline; checks what to do with it.

        Calls the appropriate sub-commands, etc.

        Parameters
        ----------
        cmdline : str
            Not the actual complete cmdline, but the cmdline after the argument
            that triggered this command.
        """
        if not cmdline.strip() or cmdline.strip().startswith('-'):
            self(cmdline.strip())
            return
        c: List[Command] = list(filter(lambda c: c.matches(cmdline), self.cmds))
        if not c:
            self.unknown_key(self, cmdline)
            return
        args = re.split(self.sep, cmdline)[1:]
        if isinstance(c[0], CascCommand):
            c[0].parse(' '.join(args))
            return
        c[0]([a.replace('"', '') for a in args])

    @classmethod
    def _check_cmd_validity(cls, cmd: Command) -> None:
        if any(map(lambda a: a.startswith('-'), cmd.aliases)):
            raise Exception('Command is not allowed to start with `-` (might be confused with argparse argument)... ')

    def add_cmd(self, cmd: Command) -> None:
        """Adds a command to the sub-commands list."""
        CascCommand._check_cmd_validity(cmd)
        cmd.parent = self
        self.cmds.append(cmd)

    def usage(self) -> str:
        """Returns the trace + options to use with this command."""
        t: str = self.trace()
        return '{} [{}]'.format(t, '|'.join([c.trace().replace(t+' ', '') for c in self.cmds]))



"""Miscellaneous functions for using pash"""

import shutil, random, sys, os
from typing import List
from enum import Enum
import colorama as cr
cr.init()

def center(txt: str) -> str:
    """Centers text appropriate for the current terminal width."""
    w = shutil.get_terminal_size().columns
    mx = max([len(l) for l in txt.split('\n')])
    return '\n'.join([' '*((w-mx)//2)+l for l in txt.split('\n')])

def print_center(txt: str) -> None:
    """Prints the centered text obtained from `center()`"""
    print(center(txt))

def printc(txt: str) -> None:
    """Alias for `print_center(txt)`"""
    print_center(txt)

def fancy_print(txt: str) -> None:
    """Prints the given text in a "fancy" way; it centers it, and gives it random splashes of colour."""
    txt = center(txt)
    col = [ cr.Fore.LIGHTRED_EX, cr.Fore.LIGHTGREEN_EX, cr.Fore.LIGHTBLUE_EX, cr.Fore.LIGHTBLACK_EX, cr.Fore.LIGHTYELLOW_EX, cr.Fore.LIGHTWHITE_EX, cr.Fore.LIGHTCYAN_EX, cr.Fore.LIGHTMAGENTA_EX, ]
    fin = ''
    while txt:
        fin += col[random.randint(0,len(col)-1)]
        r = min(random.randint(1,8), len(txt))
        fin += txt[:r]
        txt = txt[r:]
    print(fin+cr.Fore.RESET)

def fprint(txt: str) -> None:
    """Alias for `fancy_print(txt)`"""
    fancy_print(txt)

def clear() -> None:
    """Clears the console appropriate to the OS in use."""
    if sys.platform == 'win32':
        os.system('cls')
        return
    os.system('clear')

"""The possible table alignments (left, right)."""
TALIGN: Enum = Enum('TALIGN', 'LEFT RIGHT')

def print_table(tab: List[List[str]], align: TALIGN = TALIGN.RIGHT) -> None:
    """Prints the given 2d-string list in a table fashion; evenly spaced columns, etc."""
    w = shutil.get_terminal_size().columns
    ms = [max(c)+1 for c in zip(*[[len(x) for x in r] for r in tab])]
    for r in tab:
        print(' '+' '.join([('{:'+('>' if align == TALIGN.RIGHT else '')+str(ms[i])+'s}').format(c) for i, c in enumerate(r)]))

class ProgressBar(object):
    """
    Represents a cli progress bar.

    ...

    Attributes
    ----------
    max : int
        Maximum value.
    title : str
        The progress bar's title.
    val : float
        The bar's current value.
    w : int
        The bar's width in the console [characters].
    done : bool
        Whether or not, the process has finished.
    
    Methods
    -------
    update(val)
        Make some progress!
    inc(by)
        Increment the value by [by].
    end()
        End the process.
    ensure_end()
        Calls end, if the process hasn't ended according to the progress bar. 
        Should always be called to make sure the progress bar finishes.
    __refresh()
        Refresh/redraw the progress bar.
    """

    def __init__(self, max: int, title: str = '', val: float = 0.):
        """
        Parameters
        ----------
        max : int
            The maximum value.
        title : str
            The title. (optional)
        val : float
            The starting value. (optional)
        """
        self.max: int   = max
        self.title: str = title
        self.val: float = val
        self.w: int     = 20
        self.done: bool = False
        self.__refresh()

    def update(self, val: float) -> None:
        """Sets the current value to the given value"""
        if self.done:
            return
        self.val = val
        self.__refresh()
        if self.val >= self.max:
            self.end()

    def inc(self, by: float) -> None:
        """Increments the current value by the given value."""
        self.update(self.val+by)

    def end(self) -> None:
        """Ends the progress bar."""
        self.done = True
        print('\r {} ['.format(self.title)+'='*self.w+'] 100.00% ... ')

    def ensure_end(self) -> None:
        """Ensures the progress bar has ended/ends. (always call when finished!)"""
        if not self.done:
            self.end()

    def __refresh(self) -> None:
        """Refreshes the progress bar."""
        pc = int((self.val/self.max)*100)
        print('\r {} ['.format(self.title)+'='*(pc//(100//self.w))+'-'*(self.w-pc//(100//self.w))+'] {:6.2f}% ...'.format(pc), end='')


"""Template commands that are oftentimes useful"""

import pash.misc
from pash.command import Command
from typing import List

def clear(cmd: Command, args: List[str]) -> None:
    """Uses pash.misc.clear to clear the console"""
    pash.misc.clear()


import setuptools

with open("README.md", "r") as f:
    long_desc = f.read()

setuptools.setup(
    name="pash-cmd",
    version="0.0.6",
    author="MattMoony",
    author_email="m4ttm00ny@gmail.com",
    description="An interactive shell for Python scripts.",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://github.com/MattMoony/pash",
    download_url="https://github.com/MattMoony/pash/archive/v_006.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'colorama==0.4.4',
    ],
    python_requires='>=3.5',
)


"""Home to the pash shell and its default functions."""

import pash.misc as misc
from pash.command import Command, CascCommand
from typing import List, Callable, Union

def _def_unknown_cmd(cc: CascCommand, cmdline: str) -> None:
    """Default callback for unknown commands"""
    if not cmdline.strip():
        return
    print('Unknown command "%s" ... ' % cmdline.split()[0])

class Shell(CascCommand):
    """
    The class representing the actual shell the user interacts with.

    ...

    Attributes
    ----------
    prompt : str
        The shell's prompt (the string prefixing the user's input; e.g.: 'bernd@raspberrypi: ').
    __interrupt_end : bool
        Whether or not the shell should terminate when a `KeyboardInterrupt` exception (Ctrl+C) arises.
    __exited : bool
        Whether or not the shell has terminated yet.
    helpc : CascCommand
        The `help` command. It takes care of calls to 'help' and '?'.

    Methods
    -------
    exit()
        Terminates the shell. (__exited = True)
    add_cmd(cmd)
        Adds a command the terminal will respond to.
    prompt_until_exit()
        Keep asking the user for input and handling those commands, until the shell terminates.
    print_help()
        Print all possible commands and their short help summary.
    """
    def __init__(self, prompt: str = '$ ', 
                       interrupt_end: bool = False, 
                       unknown_cmd: Callable[[CascCommand, str], None] = _def_unknown_cmd,
                       sep: str = r'\s(?:(?=(?:[^"]*"[^"]*")+[^"]*$)|(?=[^"]*$))'
                ):
        """
        Parameters
        ----------
        prompt : str
            The shell's prompt.
            Default is '$ '.
        interrupt_end : bool
            Whether or not the shell should terminate on a KeyboardInterrupt.
            Default is False.
        unknown_cmd : Callable[[CascCommand, str], None]
            The function that will be called, when the shell encounters an unknown command.
            Default is _def_unknown_cmd.
        sep : str
            The argument separator.
            Default is r'\s(?:(?=(?:[^"]*"[^"]*")+[^"]*$)|(?=[^"]*$))'.
        """
        super().__init__('', unknown_key=unknown_cmd, sep=sep)
        self.__interrupt_end: bool = interrupt_end
        self.__exited: bool = False
        self.prompt: str = prompt
        self.helpc: CascCommand = CascCommand('help', '?', callback=lambda c, a: self.print_help(), hint='Displays help for any/all commands ...')
        self.add_cmd(self.helpc)

    def exit(self) -> None:
        """Terminates the shell"""
        self.__exited = True

    def add_cmd(self, cmd: Command) -> None:
        """
        Adds a command to the shell.

        ...

        Parameters
        ----------
        cmd : Command
            The command to-be-added.
        """
        Shell._check_cmd_validity(cmd)
        self.cmds.append(cmd)
        self.helpc.add_cmd(Command(*cmd.aliases, 
            case_sensitive=cmd.case_sensitive, 
            callback=lambda c, a: print(' Usage: ' + cmd.usage()+'\n Aliases: '+', '.join(cmd.aliases)+'\n Help: '+cmd.help)))

    def prompt_until_exit(self) -> None:
        """Keeps prompting the user for more input (commands) until the shell terminates."""
        self.__exited = False
        while not self.__exited:
            try:
                print(self.prompt, end='')
                cmdline: str = input()
                self.parse(cmdline)
            except KeyboardInterrupt:
                print()
                if self.__interrupt_end:
                    self.__exited = True
                    break
    
    def print_help(self) -> None:
        """Prints all available commands and their short help summaries."""
        misc.print_table([[ str(c).split('\t')[0], str(c).split('\t')[1], ] for c in self.cmds], align=misc.TALIGN.LEFT)

    def __call__(self, args: Union[str, List[str]]) -> None:
        """Does nothing - overrides parent."""
        return